import gleam/bool
import gleam/float
import gleam/int
import gleam/list
import gleam/option.{type Option, None, Some}
import gleam/pair
import gleam/result
import kicad_sexpr/parse.{type SExpr}

pub type ExprType {
  Token(name: String)
  String
  Int
  Float
  Name
  Unknown
}

pub type DecodeError {
  UnexpectedEndOfAttributes(expected: ExprType)
  UnexpectedTrailingAttributes(got: List(SExpr))
  IncorrectExprType(got: SExpr, expected: ExprType)
  IncorrectTokenName(got: String, expected: String)
  IncorrectName(got: String, expected: String)
  UnmatchedEnum(got: String, expected: List(String))
  UnmatchedIntEnum(got: Int, expected: List(Int))
  UnmatchedOneOf(got: SExpr)
  TokenAttributeError(name: String, error: DecodeError)
}

pub type Decoded(a) =
  Result(#(a, List(SExpr)), DecodeError)

pub type Decoder(a) =
  fn(List(SExpr)) -> Decoded(a)

pub type NextFn(a, b) =
  fn(a) -> Decoder(b)

pub fn token(
  named name: String,
  with decoder: Decoder(a),
  then next: NextFn(a, b),
) -> Decoder(b) {
  fn(sexprs: List(SExpr)) {
    case sexprs {
      [] -> Error(UnexpectedEndOfAttributes(Token(name)))
      [parse.Token(name: name_, attributes:), ..sexprs] if name_ == name -> {
        use #(value, rest) <- result.try(
          decoder(attributes)
          |> result.map_error(TokenAttributeError(name:, error: _)),
        )
        use <- bool.guard(
          rest != [],
          Error(TokenAttributeError(
            name:,
            error: UnexpectedTrailingAttributes(rest),
          )),
        )
        next(value)(sexprs)
      }
      [parse.Token(name: name_, ..), ..] ->
        Error(IncorrectTokenName(got: name_, expected: name))
      [sexpr, ..] -> Error(IncorrectExprType(got: sexpr, expected: Token(name)))
    }
  }
}

pub fn token_wrapper(
  named name: String,
  with attr: fn(NextFn(a, a)) -> Decoder(b),
  then next: NextFn(b, c),
) -> Decoder(c) {
  token(named: name, then: next, with: attr(success))
}

pub fn string(then next: NextFn(String, a)) -> Decoder(a) {
  fn(sexprs: List(SExpr)) {
    case sexprs {
      [] -> Error(UnexpectedEndOfAttributes(String))
      [parse.String(value), ..sexprs] -> next(value)(sexprs)
      [sexpr, ..] -> Error(IncorrectExprType(got: sexpr, expected: String))
    }
  }
}

pub fn name(then next: NextFn(String, a)) -> Decoder(a) {
  fn(sexprs: List(SExpr)) {
    case sexprs {
      [] -> Error(UnexpectedEndOfAttributes(String))
      [parse.Name(value), ..sexprs] -> next(value)(sexprs)
      [sexpr, ..] -> Error(IncorrectExprType(got: sexpr, expected: String))
    }
  }
}

pub fn name_string(then next: NextFn(String, a)) -> Decoder(a) {
  fn(sexprs: List(SExpr)) {
    case sexprs {
      [] -> Error(UnexpectedEndOfAttributes(String))
      [parse.Name(value), ..sexprs] | [parse.String(value), ..sexprs] ->
        next(value)(sexprs)
      [sexpr, ..] -> Error(IncorrectExprType(got: sexpr, expected: String))
    }
  }
}

pub fn name_number(then next: NextFn(String, a)) -> Decoder(a) {
  fn(sexprs: List(SExpr)) {
    case sexprs {
      [] -> Error(UnexpectedEndOfAttributes(String))
      [parse.Name(value), ..sexprs] -> next(value)(sexprs)
      [parse.Int(value), ..sexprs] -> next(int.to_string(value))(sexprs)
      [parse.Float(value), ..sexprs] -> next(float.to_string(value))(sexprs)
      [sexpr, ..] -> Error(IncorrectExprType(got: sexpr, expected: String))
    }
  }
}

pub fn name_number_string(then next: NextFn(String, a)) -> Decoder(a) {
  fn(sexprs: List(SExpr)) {
    case sexprs {
      [] -> Error(UnexpectedEndOfAttributes(String))
      [parse.Name(value), ..sexprs] | [parse.String(value), ..sexprs] ->
        next(value)(sexprs)
      [parse.Int(value), ..sexprs] -> next(int.to_string(value))(sexprs)
      [parse.Float(value), ..sexprs] -> next(float.to_string(value))(sexprs)
      [sexpr, ..] -> Error(IncorrectExprType(got: sexpr, expected: String))
    }
  }
}

pub fn float(then next: NextFn(Float, a)) -> Decoder(a) {
  fn(sexprs: List(SExpr)) {
    case sexprs {
      [] -> Error(UnexpectedEndOfAttributes(Float))
      [parse.Int(value), ..sexprs] -> next(int.to_float(value))(sexprs)
      [parse.Float(value), ..sexprs] -> next(value)(sexprs)
      [sexpr, ..] -> Error(IncorrectExprType(got: sexpr, expected: Float))
    }
  }
}

pub fn int(then next: NextFn(Int, a)) -> Decoder(a) {
  fn(sexprs: List(SExpr)) {
    case sexprs {
      [] -> Error(UnexpectedEndOfAttributes(Int))
      [parse.Float(value), ..sexprs] -> next(float.round(value))(sexprs)
      [parse.Int(value), ..sexprs] -> next(value)(sexprs)
      [sexpr, ..] -> Error(IncorrectExprType(got: sexpr, expected: Int))
    }
  }
}

pub fn enum(
  with variants: List(#(String, a)),
  then next: NextFn(a, b),
) -> Decoder(b) {
  fn(sexprs: List(SExpr)) {
    case sexprs {
      [] -> Error(UnexpectedEndOfAttributes(Name))
      [parse.Name(value), ..sexprs] ->
        case list.key_find(variants, value) {
          Ok(value) -> next(value)(sexprs)
          Error(Nil) ->
            Error(UnmatchedEnum(
              got: value,
              expected: list.map(variants, pair.first),
            ))
        }
      [sexpr, ..] -> Error(IncorrectExprType(got: sexpr, expected: Name))
    }
  }
}

pub fn int_enum(with variants: List(a), then next: NextFn(a, b)) -> Decoder(b) {
  fn(sexprs: List(SExpr)) {
    case sexprs {
      [] -> Error(UnexpectedEndOfAttributes(Int))
      [parse.Int(value), ..sexprs] -> {
        let variants = list.index_map(variants, fn(x, i) { #(i, x) })
        case list.key_find(variants, value) {
          Ok(value) -> next(value)(sexprs)
          Error(Nil) ->
            Error(UnmatchedIntEnum(
              got: value,
              expected: list.map(variants, pair.first),
            ))
        }
      }
      [sexpr, ..] -> Error(IncorrectExprType(got: sexpr, expected: Name))
    }
  }
}

pub fn flag(named name: String, then next: NextFn(Bool, a)) -> Decoder(a) {
  fn(sexprs: List(SExpr)) {
    case sexprs {
      [parse.Name(value), ..sexprs] if value == name -> next(True)(sexprs)
      _ -> next(False)(sexprs)
    }
  }
}

pub fn optional(
  attr: fn(NextFn(a, a)) -> Decoder(b),
  then next: NextFn(Option(b), c),
) -> Decoder(c) {
  fn(source: List(SExpr)) {
    case attr(success)(source) {
      Ok(#(value, rest)) -> next(Some(value))(rest)
      Error(TokenAttributeError(..) as err) -> Error(err)
      Error(_) -> next(None)(source)
    }
  }
}

pub fn list(
  attr: fn(NextFn(a, a)) -> Decoder(b),
  then next: NextFn(List(b), c),
) -> Decoder(c) {
  fn(source: List(SExpr)) {
    use #(value, rest) <- result.try(do_list(source, attr(success), []))
    next(value)(rest)
  }
}

fn do_list(
  source: List(SExpr),
  decoder: Decoder(a),
  results: List(a),
) -> Decoded(List(a)) {
  case decoder(source) {
    Ok(#(value, rest)) -> do_list(rest, decoder, [value, ..results])
    Error(TokenAttributeError(..) as err) -> Error(err)
    Error(_) -> Ok(#(results, source))
  }
}

pub fn of(with decoder: Decoder(a), then next: NextFn(a, b)) -> Decoder(b) {
  fn(source: List(SExpr)) {
    use #(token, rest) <- result.try(decoder(source))
    next(token)(rest)
  }
}

pub fn one_of(
  attr: fn(NextFn(a, a)) -> Decoder(b),
  or attrs: List(fn(NextFn(a, a)) -> Decoder(b)),
  then next: NextFn(b, c),
) -> Decoder(c) {
  fn(source: List(SExpr)) {
    case attr(success)(source) {
      Ok(#(value, rest)) -> next(value)(rest)
      Error(TokenAttributeError(..) as err) -> Error(err)
      Error(_) -> {
        use #(value, rest) <- result.try(do_one_of(
          source,
          list.map(attrs, fn(attr) { attr(success) }),
        ))
        next(value)(rest)
      }
    }
  }
}

fn do_one_of(source: List(SExpr), decoders: List(Decoder(a))) -> Decoded(a) {
  case decoders {
    [] ->
      Error(
        list.first(source)
        |> result.map(UnmatchedOneOf)
        |> result.unwrap(UnexpectedEndOfAttributes(Unknown)),
      )
    [decoder, ..decoders] ->
      case decoder(source) {
        Ok(#(value, rest)) -> Ok(#(value, rest))
        Error(TokenAttributeError(..) as err) -> Error(err)
        Error(_) -> do_one_of(source, decoders)
      }
  }
}

pub fn map(
  with attr: fn(NextFn(a, b)) -> Decoder(c),
  to map_fn: fn(a) -> b,
) -> fn(NextFn(c, d)) -> Decoder(d) {
  of(then: _, with: {
    use value <- attr()
    success(map_fn(value))
  })
}

pub fn success(value: a) -> Decoder(a) {
  fn(source: List(SExpr)) { Ok(#(value, source)) }
}

pub fn failure(error: DecodeError) -> Decoder(a) {
  fn(_: List(SExpr)) { Error(error) }
}

pub fn run(decoder: Decoder(a), on source: SExpr) -> Result(a, DecodeError) {
  use #(value, rest) <- result.try(decoder([source]))
  use <- bool.guard(rest != [], Error(UnexpectedTrailingAttributes(rest)))
  Ok(value)
}
